import math
import numpy as _np
import numba as _numba
import contextlib


@contextlib.contextmanager
def corrfunction(shape, z, qmax, xcenter=None, ycenter=None, threads=-1):
    """
    CPU based direct Autocorrelation with q correction

    parameters:
    shape (tuple) of inputs in pixels
    z (scalar) distance of detector in pixels
    optional
    xcenter (scalar): position of center in x direction, defaults to shape[0]/2
    ycenter (scalar): position of center in x direction, defaults to shape[1]/2
    threads (even int): cpu threads. -1 means use global setting. will always use even number of threads. will also use more memory!

    returns a function with signature float[:,:,:](float[:,:] image) that does the correlation
    """

    if threads == -1:
        parallel = True
        try:
            pmax = _numba.get_num_threads() // 2
        except Exception:  # old numba version
            pmax = min(8, _numba.config.multiprocessing.cpu_count() // 4)
    elif threads == 1 or threads == 0:
        pmax = 1
        parallel = False
    else:
        parallel = True
        pmax = parallel // 2
    xcenter = xcenter or shape[0] / 2.0
    ycenter = ycenter or shape[1] / 2.0
    y, x = _np.meshgrid(_np.arange(shape[1], dtype=_np.float64), _np.arange(shape[0], dtype=_np.float64))
    x -= xcenter
    y -= ycenter
    d = _np.sqrt(x ** 2 + y ** 2 + z ** 2)
    qx, qy, qz = [(k / d * z) for k in (x, y, z)]
    maxdqz = int(math.ceil(math.sqrt(z ** 2 - (max(qmax + 1, _np.max(qx), _np.max(qy)) - (qmax + 1)) ** 2)) - _np.min(qz)) + 1
    # the work will be split in two blocks with different qx-offset in the result,
    # because of the curvature, they have to look a bit 'backwards' to get all pais that belong to their otput part
    overscany = int(_np.ceil(_np.max(_np.max(qy, axis=0) - _np.min(qy, axis=0))))
    overscanx = int(_np.ceil(_np.max(_np.max(qx, axis=1) - _np.min(qx, axis=1))))

    del x, y, d

    def inner(input, qx, qy, qz):
        out = _np.zeros((pmax, 2 * maxdqz + 1, qmax + 1, 2 * qmax + 1), dtype=_np.float64)
        for p in _numba.prange(pmax):  # parallel threads working on diffrent cops of the output
            for d in _numba.prange(2):  # parallel threads splitting in dqx
                direction = -1 + 2 * d  # -1 +1
                for refx in range(p, shape[0], pmax):
                    for refy in range(shape[1]):
                        qxr = qx[refx, refy]
                        qyr = qy[refx, refy]
                        qzr = qz[refx, refy]
                        refv = input[refx, refy]
                        dqx = 0
                        x = max(0, min(input.shape[0] - 1, refx - direction * overscanx))
                        while -qmax <= dqx <= qmax and 0 <= x < input.shape[0]:
                            dqy = 0
                            y = max(0, refy - overscany)
                            while dqy <= qmax and 0 <= y < input.shape[1]:
                                dqy = int(round(qy[x, y] - qyr))
                                dqx = int(round(qx[x, y] - qxr))
                                dqz = int(round(qz[x, y] - qzr))
                                val = refv * input[x, y]
                                if 0 <= dqy <= qmax and d <= direction * dqx <= qmax:  # dont do dx=0 twice
                                    dqxs = dqx + qmax
                                    dqys = dqy
                                    dqzs = dqz + maxdqz
                                    out[p, dqzs, dqys, dqxs] += val
                                y += 1
                            x += direction
        return out

    finner = _numba.njit(inner, parallel=parallel, fastmath=True).compile("float64[:,:,:,:](float64[:,:],float64[:,:],float64[:,:],float64[:,:])")

    def corr(input):
        """
        Do the correlation
        """
        if finner is None:
            raise ValueError("already closed, use within with statement")
        input = _np.asarray(input).astype(_np.float64, copy=False)
        if not all(i == s for (i, s) in zip(input.shape, shape)):
            raise ValueError("input has not the shape specified in init!")
        tmp = _np.sum(finner(input, qx, qy, qz), axis=0)  # summing over the threads
        return _np.concatenate((_np.flip(tmp[:, :, :], axis=(0, 1, 2)), tmp[:, 1:, :]), axis=1)[:, :, :]  # unwapping

    yield corr
    qx = qy = qz = finner = shape = None
    for x in locals():
        del x



__all__ = [
    "hitcor",
    "hitcorrad",
    "cpucor",
    "cpucorrad",
    "cucor",
    "cucorrad",
    "cpusimple",
    "cusimple",
    "ft",
    "singleshotnorm",
]

import numba.cuda as _nbcuda
import warnings as _w

_w.filterwarnings("ignore", message="numpy.dtype size changed")
_w.filterwarnings("ignore", message="numpy.ufunc size changed")


from . import ( # noqa
    hitcor,
    hitcorrad,
    cpucor,
    cpucorrad,
    cpusimple,
    common,
    singleshotnorm,
    ft
)

# cuda
_cuda = _nbcuda.is_available()
if _cuda:
    try:
        from . import cucor, cucorrad, cusimple

        simple = cusimple
        qcor = cucor
        qcorrad = cucorrad
    except Exception as e:
        print(e)
        _cuda = False
if not _cuda:
    _w.warn("no cuda available", stacklevel=2)
    simple = cpusimple
    qcor = cpucor
    qcorrad = cpucorrad



__all__ = ["cpu", "simobj", "cuda", "common", "time", "cutime", "simple"]
import warnings as _w

from . import cpu, simobj, time, simple
from .common import *

_cuda = False

try:
    import cupy as _cp
    import numba.cuda as _numba_cuda

    if _cp.cuda.is_available() and _numba_cuda.is_available():
        from . import cuda
        from . import cutime
        auto = cuda
        autotime = cutime
        _cuda = True

except ImportError as _e:
    if not any(x in str(_e.args[0]).lower() for x in ["cuda", "libcu", "cupy"]):
        print(_e)
except Exception as _e:
    print("Import exception:")
    print(_e)


if not _cuda:
    _w.warn("cuda error. cuda time simulation not imported. is cuda available and paths set?")
    auto = cpu
    autotime = time



import unittest
import numpy as np
import numpy.testing as testing
import numba.cuda


class basic(unittest.TestCase):
    def test_import(self):
        import warnings
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            import idi.simulation


class simobject(unittest.TestCase):
    def test_sphere(self):
        from idi.simulation import simobj

        o = simobj.sphere(E=1, N=1e3, r=10)
        o.rndPos = False
        o.rndPhase = False

        self.assertAlmostEqual(o.k, 2 * np.pi / 1.24)
        self.assertEqual(o.N, 1000)

        t = o.get()
        self.assertTupleEqual(t.shape, (1000, 4))
        pos0 = t[:, :3]
        phase0 = t[:, 3:]
        testing.assert_array_less(np.linalg.norm(pos0, axis=1), 10)
        testing.assert_allclose(phase0, 0)

        o.r = 5
        t = o.get()
        self.assertTupleEqual(t.shape, (1000, 4))
        pos1 = t[:, :3]
        r = np.linalg.norm(pos1, axis=1)
        self.assertAlmostEqual(r.max(), 5, delta=0.5)
        testing.assert_array_less(r, 5)

        o.rndPhase = True
        pos2, phase2 = o.get2()
        self.assertFalse(np.allclose(phase2, 0))
        self.assertAlmostEqual(phase2.mean(), np.pi, delta=0.2)
        testing.assert_allclose(pos1, pos2)

        o.rndPos = True
        pos3, phase3 = o.get2()
        self.assertFalse(np.allclose(pos2, pos3))
        self.assertFalse(np.allclose(phase2, phase3))

        with self.assertWarns(Warning):
            o.rotangles = [1, 1, 1]

    def test_sc(self):
        from idi.simulation import simobj

        o = simobj.sc(1, 100, 1)
        pos = o.get()
        self.assertEqual(len(pos), 100)
        self.assertTrue(all([np.allclose(np.unique(c), np.arange(-2, 3)) for c in pos[:, :3].T]))

        o = simobj.sc(1, 1000, 1, repeats=[10, 10, 10])
        p0 = o.get2()[0]
        r0 = sorted(np.linalg.norm(p0, axis=1))
        o.rotangles = [0.5, 0.1, 0.1]
        p1 = o.get2()[0]
        r1 = sorted(np.linalg.norm(p1, axis=1))
        self.assertFalse(np.allclose(p0, p1))
        testing.assert_allclose(r0, r1)

        with self.assertWarns(UserWarning):
            o = simobj.sc(1, 100000, 2, fwhm=[10, 10, 10], rho=[2, 2, 10])
        self.assertGreaterEqual(np.max(o.get2()[0]), 5)
        self.assertLessEqual(np.max(o.get2()[0]), 20)

    def test_gauss(self):
        from idi.simulation import simobj

        o = simobj.gauss(1, 100, 1)
        self.assertEqual(len(o.get()), 100)

    def test_multisphere(self):
        from idi.simulation import simobj

        o = simobj.multisphere(1, 100, 1, Nspheres=10)
        self.assertEqual(len(o.get()), 100)

        o = simobj.multisphere(1, 100, 1, fwhm=10)
        self.assertEqual(len(o.get()), 100)

    def test_grating(self):
        from idi.simulation import simobj

        o = simobj.grating(1, 100, 1, 10, fwhm=[10, 10, 1], rho=[2, 2, 100], rotangles=[np.pi, 0, 0])
        self.assertEqual(len(o.get()), 100)


class cpu(unittest.TestCase):
    def setUp(self):
        from idi.simulation import simobj

        self.obj = simobj.sphere(100, 100, 10)

    def test_cpu(self):
        from idi.simulation import cpu

        res = cpu.simulate(2, self.obj, 4, 10, 1000, self.obj.k)
        self.assertTupleEqual(res.shape, (2, 4, 4))

    def test_time(self):
        from idi.simulation import time

        res = time.simulate(self.obj, 4, 10, 1000, self.obj.k, 1.0, 1.0, 1.0)
        self.assertTupleEqual(res.shape, (4, 4))

    def test_simple(self):
        from idi.simulation import simple

        res = simple.simulate(self.obj, 4, 10, 1000, self.obj.k)
        self.assertTupleEqual(res.shape, (4, 4))


@unittest.skipIf(not numba.cuda.is_available(), 'no cuda available')
class gpu(unittest.TestCase):
    def setUp(self):
        from idi.simulation import simobj

        self.obj = simobj.sphere(100, 100, 10)

    def test_cuda(self):
        from idi.simulation import cuda

        res = cuda.simulate(2, self.obj, 4, 10, 1000, self.obj.k)
        self.assertTupleEqual(res.shape, (2, 4, 4))

    def test_cutime(self):
        from idi.simulation import cutime

        res = cutime.simulate(self.obj, 4, 10, 1000, self.obj.k, 1.0, 1.0, 1.0)
        self.assertTupleEqual(res.shape, (4, 4))


if __name__ == '__main__':
    unittest.main()



import numpy as _np
import numba as _numba
from .common import _getidx

pmax = 16  # todo
debug = True


@_numba.njit()
def _ks(xi, yi, z, center):
    x = xi.astype(_numba.float64) - center[0]
    y = yi.astype(_numba.float64) - center[1]
    d = _np.sqrt(x ** 2 + y ** 2 + z ** 2)
    kx = x / d * z
    ky = y / d * z
    kz = z / d * z
    return kx, ky, kz


@_numba.njit()
def _qsize(Nx, Ny, z):
    qlenx = int(Nx)
    qleny = int(Ny * 2)
    qlenz = 2 * int((z - (z ** 2 / _np.sqrt((Nx / 2 + 1) ** 2 + (Ny / 2 + 1) ** 2 + z ** 2))) + 1)
    return qlenx, qleny, qlenz


@_numba.njit(parallel=False, boundscheck=not debug)
def _corr(input, z):
    Nx, Ny = input.shape
    xi, yi = _np.where(input > 0)
    Nhits = len(xi)
    qlenx, qleny, qlenz = _qsize(Nx, Ny, z)
    tmp = _np.zeros((qlenx, qleny, qlenz), dtype=_np.uint64)
    center = (Nx / 2.0, Ny / 2.0)
    kx, ky, kz = _ks(xi, yi, z, center)

    for n in range(Nhits):
        for m in range(n):
            qz = int(_np.rint((kz[n] - kz[m])))
            qx = int(_np.rint((kx[n] - kx[m])))
            qy = int(_np.rint((ky[n] - ky[m])))
            if qx < 0:
                qz = -qz
                qy = -qy
                qx = -qx
            qz = qz + qlenz // 2
            qy = qy + qleny // 2
            if 0 <= qx < qlenx and 0 <= qy < qleny and 0 <= qz < qlenz:
                tmp[qx, qy, qz] += input[xi[n], yi[n]] * input[xi[m], yi[m]]
    return tmp


@_numba.njit(parallel=True, fastmath=True, boundscheck=not debug)
def _pcorr(input, z):
    Nx, Ny = input.shape
    xi, yi = _np.where(input > 0)
    Nhits = len(xi)
    qlenx, qleny, qlenz = _qsize(Nx, Ny, z)
    tmp = [_np.zeros((qlenx, qleny, qlenz), dtype=_np.float64) for p in range(pmax)]
    center = (Nx / 2.0, Ny / 2.0)
    kx, ky, kz = _ks(xi, yi, z, center)

    for p in _numba.prange(pmax):
        ptmp = tmp[p]
        idx = _getidx(p, pmax, Nhits)
        for n in idx:
            for m in range(n):
                qz = int(round((kz[n] - kz[m])))
                qx = int(round((kx[n] - kx[m])))
                qy = int(round((ky[n] - ky[m])))
                if qx < 0:
                    qz = -qz
                    qy = -qy
                    qx = -qx
                qz = qz + qlenz // 2
                qy = qy + qleny // 2
                if 0 <= qx < qlenx and 0 <= qy < qleny and 0 <= qz < qlenz:
                    ptmp[qx, qy, qz] += input[xi[n], yi[n]] * input[xi[m], yi[m]]
    out = _np.zeros((qlenx, qleny, qlenz))

    for p in range(pmax):
        out += tmp[p]
    return out


def corr(input, z):
    """
    correlation of hits
    in NxN array or nx(NxN) arrays
    """
    if input.ndim == 2:
        return _pcorr(input, z)
    elif input.ndim == 3:
        return _np.sum(_pcorrs(input, z), axis=0)
    else:
        raise TypeError


@_numba.njit(parallel=True, boundscheck=not debug)
def _pcorrs(input, z):
    Nx, Ny = input.shape[-2:]
    Nimg = input.shape[0]
    qlenx = int(Nx)
    qleny = int(2 * Ny)
    qlenz = 2 * int((z - (z ** 2 / _np.sqrt((Nx / 2 + 1) ** 2 + (Ny / 2 + 1) ** 2 + z ** 2))) + 1)
    out = _np.zeros((pmax, qlenx, qleny, qlenz), dtype=_np.uint64)
    for p in _numba.prange(pmax):
        # Nperp = int(_np.floor(Nimg / pmax))
        # for n in range(p * Nperp, (p + 1) * Nperp):
        for n in range(p, Nimg, pmax):
            out[p] += _corr(input[n, ...], z)
    return out


def unwrap(input):
    ret = _np.zeros((2 * input.shape[0], input.shape[1], input.shape[2]))
    ret[1 : input.shape[0] + 1, 1:, 1:] = input[::-1, :0:-1, :0:-1]
    ret[input.shape[0] :, :, :] += input
    return ret


# should give a pattern after unwrapping...
# tst=np.zeros((5,10,12))
# tst[::2,:,tst.shape[2]//2]+=1
# tst[:,::2,tst.shape[2]//2]+=1
# tstu=unwrap(tst)
# plt.matshow(tstu[:,:,6],vmin=0,vmax=2)



import unittest

import numpy as np
import numpy.testing as testing


class basic(unittest.TestCase):
    def test_import(self):
        import idi.util  # noqa

    def test_angles(self):
        from idi.util import angles, rotation

        testing.assert_allclose(angles(np.eye(3)), (0, 0, 0), atol=1e-15)
        testing.assert_allclose(rotation(0, 0, 0), np.eye(3), atol=1e-15)
        for case in [(np.pi, 0, 0), (0, np.pi, 0), (0, 0, np.pi), (1, 1, 1)]:
            rm = rotation(*case)
            testing.assert_allclose(rotation(*angles(rm)), rm, atol=1e-15, err_msg=f"{case}: rotation(*angles({rm})) != {rm}")

    def test_fastlen(self):
        from idi.util import fastlen

        for x in [1, 2, 10, 17, 101, 5000]:
            self.assertGreaterEqual(fastlen(x), x)
            self.assertGreaterEqual(fastlen(x, factors=[2]), x)
            self.assertGreaterEqual(fastlen(x, factors=[2, 3]), x)

        xs = np.array([1, 11, 5001])
        for x, f in zip(xs, fastlen(xs, factors=[2, 3])):
            self.assertGreaterEqual(f, x)
        for x, f in zip(xs, fastlen(xs)):
            self.assertGreaterEqual(f, x)

        self.assertEqual(19, fastlen(19, [2, 19]))
        self.assertEqual(24, fastlen(19, [2, 3]))
        self.assertEqual(32, fastlen(19, [2]))
        self.assertEqual(6048, fastlen(6001, factors=[2, 3, 5, 7, 11]))

    def test_radial_profile(self):
        from idi.util import radial_profile

        # basic
        t = np.ones((100, 100))
        r, s = radial_profile(t, calcStd=True)
        testing.assert_allclose(r, 1, atol=1e-15)
        testing.assert_allclose(s, 0, atol=1e-15)
        self.assertEqual(len(r), 72)
        self.assertEqual(True, True)
        # center
        t = np.zeros((10, 10))
        t[2:5, :3] = 1
        r = radial_profile(t, center=(3, 1))
        testing.assert_allclose(r, np.array([1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]))
        # random
        t = np.random.default_rng(0).random((50, 50))
        r, s = radial_profile(t, calcStd=True)
        testing.assert_array_less(r, 1)
        testing.assert_array_less(0, r)
        testing.assert_array_less(0, s[1:])
        testing.assert_allclose(r.mean(), 0.5, atol=0.1)
        # nan
        t[t > 0.8] = np.nan
        t[~np.isnan(t)] = 1
        r = radial_profile(t)
        testing.assert_allclose(r, 1, atol=1e-15)
        self.assertFalse(np.any(np.isnan(r)))

    def test_abs(self):
        from idi.util import abs2, abs2c

        self.assertAlmostEqual(abs2(1j + 1), 2)
        t = np.array(1 - 1j, dtype=np.complex128)
        self.assertAlmostEqual(abs2c(t), 2.0)
        self.assertFalse(np.iscomplex(abs2c(t)), 2.0)
        self.assertTrue(np.iscomplexobj(abs2c(t)), 2.0)

    def test_shortsci(self):
        from idi.util import shortsci

        for n, s in zip(
            [1, 0.1, -0.1, 10, 1e17, -1e-17, 0.5, 1.5, -15],
            ["1e0", "1e-1", "-1e-1", "1e1", "1e17", "-1e-17", "5e-1", "2e0", "-2e1"],
        ):
            self.assertEqual(shortsci(n), s)

        for n, s in zip([1, 0.555, 1.51, -1.234e17], ["1.0e0", "5.6e-1", "1.5e0", "-1.2e17"]):
            self.assertEqual(shortsci(n, decimals=1), s)


class random(unittest.TestCase):
    """
    TODO:
    poisson_disc_sample
    rndConvexpoly
    rndIcosahedron
    rndstr
    gnorm
    """

    def test_rndSphere(self):
        from idi.util import rndSphere

        R = 10
        N = int(1e6)
        t = rndSphere(R, N)
        r = np.linalg.norm(t, axis=1)
        self.assertTupleEqual(t.shape, (N, 3))
        self.assertGreater(r.max(), R * 0.95)
        self.assertLessEqual(r.max(), R)
        hist = np.histogram(r, bins=np.linspace(0, 1, int(N ** (1 / 3) * 0.2)) ** (1 / 3) * R)[0]
        self.assertLess(np.ptp(hist) / np.mean(hist), 0.05, msg="r deviation higher than 5%")
        d, b = np.histogramdd(t, density=True, bins=11)
        br = np.linalg.norm(np.meshgrid(*((c[:-1] + c[1:]) / 2 for c in b)), axis=0)
        f = d[br < 0.9 * R]
        self.assertLess(f.std() / f.mean(), 0.05, msg="density deviation higher than 5%")

    def test_rndstr(self):
        from idi.util import rndstr

        self.assertEqual(len(rndstr(0)), 0)
        self.assertEqual(len(rndstr(1)), 1)
        self.assertEqual(len(rndstr(1000)), 1000)
        s = set()
        for i in range(100):
            c = rndstr(5)
            s.add(c)
            self.assertTrue(c.isascii())
        self.assertEqual(len(s), 100, msg="repeats found")

    def test_rndgennorm(self):
        from idi.util import rndgennorm, fwhm

        fwhms = (1, 5, 10)
        t = rndgennorm(0, fwhms, (2, 2, 100), int(1e7))
        np.testing.assert_allclose(np.mean(t, axis=0), 0, atol=0.01, err_msg="mean failed")
        self.assertTupleEqual(t.shape, (int(1e7), 3), msg="shape missmatch")
        hist, bins = zip(*(np.histogram(x, bins=int(1e3), range=(-10, 10)) for x in t.T))
        for h, b, f in zip(hist, bins, fwhms):
            self.assertAlmostEqual(fwhm(b, h), f, delta=0.1, msg="fwhm failed")

    def test_randomrotation(self):
        from idi.util import random_rotation

        t = random_rotation()
        self.assertTupleEqual(t.shape, (3, 3))
        self.assertAlmostEqual(abs(np.linalg.det(t)), 1)


class h5util(unittest.TestCase):
    def setUp(self):
        import h5py

        self.f1 = h5py.File(name="f1", driver="core", backing_store=False, mode="w")
        self.f1.create_group("group")

    def tearDown(self):
        self.f1.close()

    def test_append(self):
        from idi.util import h5util

        h5util.appenddata(self.f1, "t1", np.zeros((1, 10, 10)))
        t = self.f1["t1"]
        self.assertTupleEqual(t.shape, (1, 10, 10))
        self.assertRaises(KeyError, h5util.appenddata, t, "blub", np.zeros((3, 10, 10)))
        h5util.appenddata(t, "/", np.zeros((3, 10, 10)))
        self.assertTupleEqual(t.shape, (4, 10, 10))
        self.assertEqual(t.compression, "lzf")
        self.assertEqual(t.shuffle, True)
        self.assertTupleEqual(t.chunks, (1, 10, 10))

        h5util.appenddata(self.f1, "string", "s1")
        h5util.appenddata(self.f1, "string", "s2_ismuchlonger")
        h5util.appenddata(self.f1, "string", np.array(self.f1["string"]))
        t = self.f1["string"]
        self.assertEqual(len(t), 4)
        self.assertEqual(t[-1], b"s2_ismuchlonger")
        self.assertEqual(np.array(t)[-2], b"s1")

        h5util.appenddata(self.f1["group"], "t2", np.ones(5))
        self.assertTupleEqual(self.f1["group/t2"].shape, (5,))

    def test_overwrite(self):
        from idi.util import h5util

        h5util.overwritedata(self.f1, "t1", np.zeros((1, 10, 10)))
        t = self.f1["t1"]
        self.assertTupleEqual(t.shape, (1, 10, 10))
        self.assertEqual(t.compression, "lzf")
        self.assertEqual(t.shuffle, True)
        self.assertTupleEqual(t.chunks, (1, 10, 10))

        h5util.overwritedata(self.f1, "t1", np.zeros((5, 10)), chunks=(2, 1))
        t = self.f1["t1"]
        self.assertTupleEqual(t.shape, (5, 10))
        self.assertTupleEqual(t.chunks, (2, 1))

        self.assertRaises(TypeError, h5util.overwritedata, t, "blub", np.zeros((3, 10, 10)))

        h5util.overwritedata(self.f1, "t2", "test")
        self.assertEqual(self.f1["t2"][0], b"test")

        h5util.overwritedata(self.f1, "t2", "othertest")
        self.assertEqual(self.f1["t2"][0], b"othertest")

        h5util.overwritedata(self.f1["group"], "t3", np.array(["a", "test"]))
        self.assertEqual(self.f1["group/t3"][1], b"test")

    def test_chunkediter(self):
        from idi.util import h5util

        self.f1["t4"] = np.zeros((15, 10, 1))

        for i, el in enumerate(h5util.chunkediter(self.f1["t4"], outsize=2)):
            if i < 7:
                self.assertTupleEqual(el.shape, (2, 10, 1), msg=f"differ at {i}: {el.shape}")
            else:
                self.assertTupleEqual(el.shape, (1, 10, 1), msg=f"differ at {i}: {el.shape}")
        self.assertEqual(i, 7)

        for i, el in enumerate(h5util.chunkediter(self.f1["t4"], outsize=1)):
            self.assertTupleEqual(el.shape, (10, 1), msg=f"differ at {i}: {el.shape}")
        self.assertEqual(i, 14)

        for i, el in enumerate(h5util.chunkediter(self.f1["t4"], outsize=5)):
            self.assertTupleEqual(el.shape, (5, 10, 1), msg=f"differ at {i}: {el.shape}")
        self.assertEqual(i, 2)

        # h5util.overwritedata(self.f1, 'string', 's1')
        # h5util.overwritedata(self.f1, 'string', 's2_ismuchlonger')
        # h5util.overwritedata(
        #     self.f1, 'string', np.array(self.f1['string']),
        # )
        # t = self.f1['string']
        # self.assertEqual(len(t), 4)
        # self.assertEqual(t[-1], b's2_ismuchlonger')
        # self.assertEqual(np.array(t)[-2], b's1')


class funchelper(unittest.TestCase):
    def test_asgen(self):
        from idi.util import asgen
        import types

        @asgen
        def test(el):
            return el

        t = test([1, 2, 3])
        self.assertIsInstance(t, types.GeneratorType)
        self.assertEqual(list(t), [1, 2, 3])

    def test_group(self):
        from idi.util import group

        a = range(10)
        for value, truth in zip(group(a, 4), [tuple(range(n, n + 4)) for n in range(0, 20, 4)]):
            self.assertTupleEqual(value, truth)


class array(unittest.TestCase):
    """
    TODO:
    create_mask
    diffdist
    fftfilter_mean
    fftfilter_std
    filter_std
    arrayfromiter
    split
    """

    def test_cutnan(self):
        from idi.util import cutnan

        t = np.zeros((10, 20))
        t[:5, :] = np.nan
        t[-5, :] = np.nan
        t[:, :2] = np.nan
        t[:, 15:17] = np.nan
        self.assertEqual(cutnan(t).shape, (4, 16))

        t = np.zeros((1, 20))
        t[:] = np.nan
        self.assertEqual(cutnan(t).shape, (0, 0))

    def test_centeredpart(self):
        from idi.util import centered_part

        t = np.arange(80).reshape(4, 4, 5)
        testing.assert_array_equal(centered_part(t, (1, 2, 2)), np.array([[[47, 48], [52, 53]]]))

    def test_list2array(self):
        from idi.util import list2array

        self.assertRaises(TypeError, list2array, 1)

        li = [1, 2, 3]
        testing.assert_array_equal(list2array(li), np.array([1, 2, 3]))

        li = [[1], [1, 2], [1, 2, 3], [1]]
        a = list2array(li)
        self.assertEqual(a[0, 0], li[0][0])
        self.assertEqual(a[1, 1], li[1][1])
        self.assertEqual(a[2, 2], li[2][2])
        self.assertEqual(np.sum(a), 11)
        self.assertEqual(len(a), len(li))

        li = [[1.0], ["a", "b"], ["a", "b", "c"], ["a"]]
        a = list2array(li)
        self.assertEqual(float(a[0, 0]), li[0][0])
        self.assertEqual(a[1, 1], li[1][1])
        self.assertEqual(a[2, 2], li[2][2])
        self.assertEqual(len(a), len(li))

    def test_fill(self):
        from idi.util import fill

        t = np.arange(20).reshape(4, 5).astype(float)
        t[1:4, 1:4] = 1
        t[2, 2] = np.nan
        t[0, 0] = np.nan
        t[-1, 0] = np.nan
        f = fill(t)
        good = np.copy(t)
        good[2, 2] = 1
        good[0, 0] = 5
        good[-1, 0] = 10
        self.assertFalse(np.any(np.isnan(f)))
        testing.assert_array_equal(good, f)

        t = np.empty((3, 3))
        t[:] = np.nan
        f = fill(t)
        self.assertTrue(np.all(np.isnan(f)))

        t = np.empty((3, 3))
        t[:] = np.nan
        t[1, 1] = 1
        f = fill(t)
        testing.assert_array_equal(f, 1)

    def test_atleastnd(self):
        from idi.util import atleastnd

        self.assertTupleEqual(atleastnd(0, 2).shape, (1, 1))
        self.assertTupleEqual(atleastnd(np.zeros((2, 2)), 1).shape, (2, 2))
        self.assertTupleEqual(atleastnd(np.zeros((3, 2)), 4).shape, (1, 1, 3, 2))

    def test_rebin(self):
        from idi.util import rebin

        t = np.arange(20).reshape(4, 5).astype(float)
        testing.assert_array_equal(rebin(t, (1, 1), "sum"), t)
        testing.assert_array_equal(rebin(t, (2, 2), "sum"), np.array([[12.0, 20.0], [52.0, 60.0]]))
        self.assertEqual(rebin(t, 4, "sum"), t[:, :4].sum())
        self.assertEqual(rebin(t, 3, "max"), t[:3, :3].max())
        self.assertEqual(rebin(t, 5, "max"), t.max())

        t = np.ones((4, 4), bool)
        self.assertEqual(rebin(t, 2, "min").dtype, bool)
        testing.assert_array_equal(
            rebin(
                t,
                2,
            ),
            np.ones((2, 2)),
        )

        t = np.ones((20, 2, 2))
        testing.assert_array_equal(rebin(t, (10, 20, 1), "mean"), np.ones((2, 1, 2)))

        self.assertRaises(ValueError, rebin, t, 1, "blub")
        self.assertRaises(ValueError, rebin, t, (1, 1), "sum")

    def test_indextreme(self):
        from idi.util import indmax, indmin

        t = np.array([1, 0, 1, 2, 1, np.nan])
        self.assertTupleEqual(indmax(t), (3,))
        self.assertTupleEqual(indmin(t), (1,))

        t = np.ones((4, 4, 4))
        t[1, 2, 3] = 2
        t[3, 2, 1:] = 0
        self.assertTupleEqual(indmax(t), (1, 2, 3))
        self.assertTupleEqual(indmin(t), (3, 2, 1))

    def test_intersect2d(self):
        from idi.util import intersect2d

        a = np.array([[1, 2, 3], [2, 3, 4], [1, 2, 3], [9, 9, 9]])
        b = np.array([[3, 2, 1], [4, 3, 2], [3, 2, 1], [9, 9, 9]])
        c = intersect2d(a, b)
        self.assertTupleEqual(c.shape, (1, 3))
        self.assertTupleEqual(tuple(c.ravel()), (9, 9, 9))

    def test_alignedarray(self):
        from idi.util import alignedarray

        t = alignedarray((3,), dtype=np.float32, alignment=64)
        self.assertEqual(t.shape, (3,))
        self.assertEqual(t.ctypes.data % 64, 0)

        t = alignedarray((2, 17), alignment=128, zero=True)
        self.assertEqual(t.shape, (2, 17))
        self.assertEqual(t.ctypes.data % 128, 0)


class accum(unittest.TestCase):
    def test_basic(self):
        from idi.util import accumulator

        a = accumulator(False)
        self.assertEqual(None, a.mean)
        self.assertEqual(None, a.std)
        self.assertEqual(a.n, 0)

        a.add(np.ones((2, 10, 10)))
        t = np.ones((2, 10, 10))
        t[0, ...] = 0
        a.add(t)
        self.assertEqual(a.n, 2)
        testing.assert_allclose(a.mean, np.array((0.5, 1))[:, None, None] * np.ones((2, 10, 10)))
        testing.assert_allclose(a.std, np.array((0.5, 0))[:, None, None] * np.ones((2, 10, 10)))
        testing.assert_allclose(a.result.mean, np.array((0.5, 1))[:, None, None] * np.ones((2, 10, 10)))
        testing.assert_allclose(a.result.std, np.array((0.5, 0))[:, None, None] * np.ones((2, 10, 10)))

    def test_shape(self):
        from idi.util import accumulator

        a = accumulator(False)
        a.add(np.ones((2, 10, 10)))
        self.assertTupleEqual(a.shape, (2, 10, 10))
        a.add(np.ones((1, 10, 1)))
        a.add(np.ones((1, 1, 1)))
        testing.assert_allclose(a.result.mean, np.ones((2, 10, 10)))
        self.assertRaises(ValueError, a.add, np.ones((3, 3, 3)))

    def test_minmax(self):
        from idi.util import accumulator

        a = accumulator(True)
        a.add(np.ones(3))
        a.add(np.array((0, 1, 2)))
        self.assertEqual(a.n, 2)
        testing.assert_allclose(a.result.mean, np.array((0.5, 1, 1.5)))
        testing.assert_allclose(a.result.std, np.array((0.5, 0, 0.5)))
        testing.assert_allclose(a.max, np.array((1, 1, 2)))
        testing.assert_allclose(a.min, np.array((0, 1, 1)))

    def test_combine(self):
        from idi.util import accumulator

        a = accumulator(True)
        a.add(np.ones(2))

        b = accumulator(False)
        b.add(np.ones(2))
        self.assertRaises(ValueError, a.combine, b)

        b = accumulator(True)
        b.add(np.ones(3))
        self.assertRaises(ValueError, a.combine, b)

        b = accumulator(True)
        b.add(np.zeros(2))
        a.combine(b)
        testing.assert_allclose(a.std, 0.5)
        testing.assert_allclose(a.mean, 0.5)
        testing.assert_allclose(a.max, 1)
        testing.assert_allclose(a.min, 0)
        self.assertEqual(a.n, 2)


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
import numpy.testing as testing
import numba.cuda


class basic(unittest.TestCase):
    def test_import(self):
        import warnings

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            import idi.reconstruction


class ft_test(unittest.TestCase):
    def test_correlator(self):
        from idi.reconstruction import ft

        with ft.correlator(np.zeros((32, 32), bool), 1000) as correlator:
            t = np.ones((32, 32))
            c = correlator.unwrap(correlator.corr(t))
            self.assertTupleEqual(c.shape, (1, 64, 64))
            self.assertTupleEqual(np.unravel_index(np.argmax(c), c.shape), (0, 32, 32))
            self.assertAlmostEqual(c.max(), 32 * 32)

        with ft.correlator(np.zeros((32, 16), bool), 100) as correlator:
            t = np.ones((32, 16))
            c = correlator.unwrap(correlator.corr(t))
            self.assertTupleEqual(c.shape[1:], (32, 64))
            self.assertTupleEqual(np.unravel_index(np.argmax(c), c.shape), (0, 16, 32))

        mask = np.zeros((17, 27), bool)
        mask[5:10] = True
        with ft.correlator(mask, 64) as correlator:
            t = np.ones((17, 27))
            c1 = correlator.unwrap(correlator.corr(t))
            t[mask] = 100
            c2 = correlator.unwrap(correlator.corr(t))
            testing.assert_allclose(c1, c2)

    def test_tiles_correlator(self):
        from idi.reconstruction import ft
        import itertools

        def _qs(shape, z, offset):
            y, x = np.meshgrid(np.arange(shape[1], dtype=np.float64), np.arange(shape[0], dtype=np.float64))
            x -= offset[0]
            y -= offset[1]
            d = np.sqrt(x**2 + y**2 + z**2)
            return np.array([(k / d * z) for k in (z, y, x)])

        z = 16
        N0, N1 = 8, 8
        off = 1
        qs = np.stack([_qs((N0, N1), z, (i, j)).T for i, j in itertools.product([-off, N0 + off], [-off, N1 + off])]).reshape(
            4, -1, 3
        )
        mean = np.ones((4, N0 * N1))
        maskout = np.zeros((4, N0 * N1), bool)

        with ft.correlator_tiles(qs, maskout, mean) as correlator:
            correlator.add(mean)
            correlator.add(mean)
            c = correlator.result(True)
            self.assertTupleEqual(c.shape, (4, 32, 32))
            self.assertAlmostEqual(np.nanmax(c), 1)
            self.assertAlmostEqual(np.nanmin(c), 1)
            self.assertGreater(np.nansum(c), N0 * N1)

    def test_corr(self):
        from idi.reconstruction import ft

        c = ft.corr(np.ones((17, 8)), 16)
        self.assertTupleEqual(c.shape[1:], (16, 32))

    def test_unwrap(self):
        from idi.reconstruction import ft

        pass


@unittest.skipIf(not numba.cuda.is_available(), "no cuda available")
class gpu_test(unittest.TestCase):
    def test_cucor(self):
        from idi.reconstruction import cucor

        t = np.ones((32, 32))
        with cucor.corrfunction(t.shape, 1000, 32) as f:
            c = f(t)
            self.assertTupleEqual(c.shape[1:], (64, 64))
            self.assertGreater(len(c), 1)
            self.assertEqual(c.max(), t.sum())
            maxid = np.unravel_index(np.argmax(c), c.shape)
            self.assertTupleEqual(maxid, tuple((i / 2 for i in c.shape)))

    def test_cucorrrad(self):
        from idi.reconstruction import cucorrad

        t = np.ones((32, 32))
        with cucorrad.corrfunction(t.shape, 100, 16) as f:
            c = f(t)
            self.assertEqual(len(c), 16)
            self.assertEqual(c.min(), t.sum())

    def test_cusimple(self):
        from idi.reconstruction import cusimple
        import scipy.signal as ss

        t = np.arange(32 * 16).reshape(32, 16).astype(float)
        c = cusimple.corr(t)
        g = ss.correlate(t, t, "full", "fft")
        testing.assert_allclose(c[1:, 1:], g, atol=1e-7)

        t = np.arange(15 * 33).reshape(15, 33).astype(float)
        c = cusimple.corr(t, norm=True)
        gc = ss.correlate(t / t.mean(), t / t.mean(), "full", "fft")
        gn = ss.correlate(np.ones_like(t), np.ones_like(t), "full", "fft")
        g = gc / gn
        g[gn < 0.9] = np.nan
        testing.assert_allclose(c[1:, 1:], g, atol=1e-7)


class cpu_test(unittest.TestCase):
    def test_cpucor(self):
        from idi.reconstruction import cpucor

        t = np.ones((32, 32))
        with cpucor.corrfunction(t.shape, 100, 16) as f:
            c = f(t)
            self.assertTupleEqual(c.shape[1:], (33, 33))
            self.assertGreater(len(c), 1)
            self.assertEqual(c.max(), t.sum())
            maxid = np.unravel_index(np.argmax(c), c.shape)
            self.assertTupleEqual(maxid, tuple((i // 2 for i in c.shape)))

    def test_cpucorrrad(self):
        from idi.reconstruction import cpucorrad

        t = np.ones((32, 32))
        with cpucorrad.corrfunction(t.shape, 100, 16) as f:
            c = f(t)
            self.assertEqual(len(c), 16)
            self.assertEqual(c.min(), t.sum())

    def test_cpusimple(self):
        from idi.reconstruction import cpusimple
        import scipy.signal as ss
        import scipy.fft

        t = np.arange(32 * 16).reshape(32, 16).astype(float)
        c = cpusimple.corr(t)
        g = ss.correlate(t, t, "full", "fft")
        testing.assert_allclose(c[1:, 1:], g, atol=1e-7)

        t = np.arange(15 * 33).reshape(15, 33).astype(float)
        c = cpusimple.corr(t, fftfunctions=[scipy.fft.rfftn, scipy.fft.irfftn], norm=True)
        gc = ss.correlate(t / t.mean(), t / t.mean(), "full", "fft")
        gn = ss.correlate(np.ones_like(t), np.ones_like(t), "full", "fft")
        g = gc / gn
        g[gn < 0.9] = np.nan
        testing.assert_allclose(c[1:, 1:], g, atol=1e-7)

        t = np.stack((t, t))
        c = cpusimple.corr(t, axes=(-2, -1), fftfunctions=[np.fft.fftn, np.fft.ifftn])[0]
        g = ss.correlate(t[0], t[0], "full", "fft")
        testing.assert_allclose(c[1:, 1:], g, atol=1e-7)

        t = t.swapaxes(0, -1)
        c = cpusimple.corr(t, axes=(0, 1))[..., 0]
        g = ss.correlate(t[..., 0], t[..., 0], "full", "fft")
        testing.assert_allclose(c[1:, 1:], g, atol=1e-7)


class singleshotnorm_test(unittest.TestCase):
    def setUp(self):
        from idi.reconstruction import singleshotnorm

        mask = np.ones((64, 64), bool)
        mask[:10, :10] = 0
        self.correlator = singleshotnorm.correlator(mask)

    def test_corr(self):
        t = np.ones((64, 64))
        c = self.correlator.corr(t)
        excepted = np.ones((127, 127))
        excepted[:10, :10] = 0
        excepted[-10:, -10:] = 0
        testing.assert_allclose(c, excepted, atol=1e-12)

    def test_properties(self):
        mask = np.ones((64, 64), bool)
        mask[:10, :10] = 0
        testing.assert_array_equal(self.correlator.mask, mask)
        self.assertTupleEqual(self.correlator.shape_input, (64, 64))
        self.assertTupleEqual(self.correlator.shape_result, (127, 127))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import os
import re
import sys
from pathlib import Path
from setuptools import setup, Extension, find_packages
from setuptools.command.sdist import sdist  # Python 3.12 compatible sdist
from Cython.Build import cythonize
import tomli
import numpy

# -----------------------------------------------------------------------------
# Heuristics for include and library directories


def find_include_and_lib_dirs():
    numpy_dir = Path(numpy.__file__).resolve()
    sys_prefix = Path(sys.prefix).resolve()

    candidate_dirs = [
        numpy_dir.parent.parent.parent.parent,
        numpy_dir.parent.parent.parent,
        sys_prefix,
    ]
    for p in sys.path:
        if p:
            parts = p.split("site-packages")
            if parts and parts[0]:
                candidate_dirs.append(Path(parts[0]) / ".." / "..")
                candidate_dirs.append(Path(parts[0]) / "..")
    for p in os.environ.get("PATH", "").split(os.pathsep):
        if p:
            candidate_dirs.append(Path(p) / "..")
    unique_basedirs = list(dict((str(p.resolve()), p.resolve()) for p in candidate_dirs).values())

    default_include_dirs, default_lib_dirs = [], []

    library_dirs = [Path(p) for p in default_lib_dirs]
    for b in unique_basedirs:
        library_dirs.append(b / "lib")
        library_dirs.append(b / "lib64")
        library_dirs.append(b / "libraries")
        library_dirs.append(b / "Library" / "lib")
        library_dirs.append(b / "Library" / "bin")
    include_dirs = [
        Path("idi"),
        Path("idi") / "reconstruction",
        Path("/usr/include"),
        Path("/usr/local/include"),
        Path("/usr/include/x86_64-linux-gnu"),
        Path("/usr/include/arm-linux-gnueabihf"),
        Path("/usr/include/i386-linux-gnu"),
        Path("/Library/Developer/CommandLineTools/SDKs/MacOSX.sdk/usr/include"),
        Path("/System/Library/Frameworks"),
    ]
    include_dirs.extend([Path(p) for p in default_include_dirs])

    for b in unique_basedirs:
        include_dirs.append(b / "include")
        include_dirs.append(b / "Library" / "include")
    library_dirs = [str(p.resolve()) for p in library_dirs if p.is_dir()]
    include_dirs = [str(p.resolve()) for p in include_dirs if p.is_dir()]
    basedirs = [str(p.resolve()) for p in unique_basedirs if p.is_dir()]

    return include_dirs, library_dirs, basedirs


# -----------------------------------------------------------------------------
# MKL detection heuristic remains mostly unchanged.


def get_mkl_info(include_dirs, library_dirs):
    found_mkl = False
    found_mkl_name = "mkl_rt"
    for d in library_dirs:
        dpath = Path(d)
        try:
            for f in dpath.iterdir():
                if f.name in {"mkl_rt.dll", "mkl_rt.so", "mkl_rt.dylib"}:
                    found_mkl = True
                    found_mkl_name = "mkl_rt"
                elif "mkl_rt.so." in f.name and not found_mkl:
                    found_mkl_name = ":" + f.name
                    found_mkl = True
            if found_mkl:
                break
        except Exception:
            continue
    libs = [found_mkl_name]
    if os.name != "nt":
        libs.extend(["pthread"])
    return [], libs


# -----------------------------------------------------------------------------
# Build the extension module.


def get_extension():
    src = Path("idi") / "reconstruction"
    sources = [str(src / "autocorrelate3.pyx")]
    print("Extension sources:", sources)
    heuristic_includes, heuristic_libs, _ = find_include_and_lib_dirs()
    include_dirs = [str(src), numpy.get_include()]
    include_dirs.extend(heuristic_includes)
    mkl_includes, mkl_libs = get_mkl_info(include_dirs, heuristic_libs)
    include_dirs.extend(mkl_includes)
    rpath = ["$ORIGIN/../../../../../../", "$ORIGIN/../../../../../", "$ORIGIN/../../../../", "$ORIGIN/../../../"]
    rpath = [f"-Wl,-rpath,{r}" for r in rpath]
    ext = Extension(
        name="idi.reconstruction.autocorrelate3",
        sources=sources,
        include_dirs=include_dirs,
        library_dirs=heuristic_libs,
        libraries=mkl_libs,
        extra_compile_args=["-DNDEBUG", "-O3", "-DMKL_ILP64"],
        extra_link_args=rpath,
    )
    return ext


def get_version(rel_path="idi/__init__.py"):
    version_file = Path(__file__).parent / rel_path
    content = version_file.read_text(encoding="utf-8")
    match = re.search(r'^__version__\s*=\s*[\'"]([^\'"]+)[\'"]', content, re.M)
    if match:
        return match.group(1)
    raise RuntimeError("Unable to find version string.")


# -----------------------------------------------------------------------------
# Read metadata from pyproject.toml.


def get_metadata():
    meta_file = Path(__file__).parent / "pyproject.toml"
    with meta_file.open("rb") as f:
        pyproject = tomli.load(f)
    project = pyproject["project"]
    return {
        "name": project["name"],
        "description": project["description"],
        "python_requires": project["requires-python"],
        "install_requires": project["dependencies"],
    }


# -----------------------------------------------------------------------------
# Setup the package.


def setup_package():
    ext = get_extension()
    ext_modules = cythonize(
        [ext],
        include_path=["idi/reconstruction"],
        compiler_directives={"language_level": "3"},
    )

    metadata = dict(
        version=get_version("idi/__init__.py"),
        packages=find_packages(),
        package_data={"": ["*.cu"]},
        cmdclass={"sdist": sdist},
        ext_modules=ext_modules,
    )
    metadata.update(get_metadata())
    setup(**metadata)


if __name__ == "__main__":
    setup_package()



import abc as _abc
from warnings import warn as _warn
import numba as _numba
import numexpr as _ne
import numpy as _np
from ..util import angles as _angles
from ..util import axisrotation as _axisrotation
from ..util import fastlen as _fastlen
from ..util import poisson_disc_sample as _pds
from ..util import random_rotation as _random_rotation
from ..util import rndgennorm as _rndgennorm
from ..util import rndIcosahedron as _rndIcosahedron
from ..util import rndSphere as _rndSphere
from ..util import rotation as _rotation
from ..util import gnorm as _gnorm
from ..util import rndtruncexp as _rndtruncexp

from numpy import pi
from gc import collect as _gc

"""
simulation objects
"""


class simobj(_abc.ABC):
    """
    baseclass
    """

    rng = _np.random.default_rng(_np.random.randint(2 << 60))
    _pos = None
    _resetproperties = []
    _rotmatrix = None

    def __init__(self, E, N):
        self.E = E  # _np.ones(self._N) * E
        self.rndPhase = True
        self.rndPos = True
        self._N = int(N)

    @property
    def N(self):
        return self._N

    @property
    def k(self):
        return 2 * pi / (1.24 / self.E)

    def get2(self):
        if self.rndPos or self._pos is None:
            self._pos = None
            # _gc()
            self.updatePos()

        if self.rndPhase:
            phase = self.rng.uniform(0, 2 * pi, (self.N, 1))
        else:
            phase = _np.zeros((self.N, 1), self._pos.dtype)
        pos = self._pos
        return pos, phase

    def get(self):
        ret = _np.empty((self.N, 4), order="C")
        _np.concatenate(self.get2(), -1, out=ret)
        return ret

    def getImg(self, dx, ndim=2):
        if not 0 < ndim <= 3:
            raise ValueError
        pos, phase = self.get2()
        ind = _np.rint((pos[:, :ndim] - _np.min(pos[:, :ndim], axis=0, keepdims=True)) / dx).astype(int)
        s = _np.array(ind.max(0)) + 1
        pads = _np.array([_fastlen(i) for i in s])
        ind += (pads - s) // 2
        img = _np.zeros(pads, _np.complex128)
        _np.add.at(img, tuple(ind.T), _np.exp(1j * _np.squeeze(phase)))
        return img

    @_abc.abstractmethod
    def updatePos(self):
        pass

    @property
    def rotangles(self):
        if self._rotmatrix is None:
            _warn("Rotation not used for this object")
            return [0, 0, 0]
        elif self._rotmatrix is False:
            return [0, 0, 0]
        else:
            return _angles(self._rotmatrix)

    @rotangles.setter
    def rotangles(self, value):
        if self._rotmatrix is None:
            _warn("Rotation not used for this object")
        elif not _np.any(value):
            self._rotmatrix = False
        elif self._rotmatrix is False:
            newmatrix = _rotation(*value)
            self._rotmatrix = newmatrix
        else:
            newmatrix = _rotation(*value)
            self._rotate(newmatrix @ self._rotmatrix.T)
            self._rotmatrix = newmatrix

    def __setattr__(self, name, value):
        # reset cached values if property changes
        if any(name == n for n in self._resetproperties):
            self._pos = None
        # _gc()
        super().__setattr__(name, value)

    def _rotate(self, rotmatrix):
        pass


class sphere(simobj):
    """
    a sphere with random positions inside
    """

    def __init__(self, E, N, r):
        self.r = r
        super().__init__(E, N)
        self._resetproperties = ["r"]

    def updatePos(self):
        self._pos = _rndSphere(self.r, self.N)


class icosahedron(simobj):
    """
    icosahedron with random positions inside
    """

    def __init__(self, E, N, r, rotangles=(0, 0, 0)):
        self.r = r
        self.rndOrientation = False
        self._rotmatrix = _rotation(*rotangles)
        super().__init__(E, N)
        self._resetproperties = ["r"]

    def updatePos(self):
        if self.rndOrientation:
            self._rotmatrix = _random_rotation(self.rng)
        self._pos = _rndIcosahedron(self.r, self._N, self._rotmatrix)

    def _rotate(self, rotmatrix):
        self._pos = _np.matmul(rotmatrix, self._pos.T, order="F").T


class gnorm(simobj):
    """
    a generalised normal shaped volume with random positions inside
    """

    def __init__(self, E, N, fwhm, rho=2, rotangles=(0, 0, 0)):
        self.fwhm, self.rho = _np.array(fwhm), _np.array(rho)
        self._rotmatrix = _rotation(*rotangles) if _np.any(rotangles) else False
        super().__init__(E, N)
        self._resetproperties = ["fwhm", "rho"]

    def updatePos(self):
        _gc()
        if _np.all(self.rho == 2):
            self._pos = self.rng.normal(scale=(self.fwhm / 2.355), size=(self.N, 3))
        else:
            self._pos = _rndgennorm(0, self.fwhm, self.rho, (self.N, 3), self.rng)
        if self._rotmatrix is not False:
            self._rotate(self._rotmatrix)

    def _rotate(self, rotmatrix):
        self._pos = _np.matmul(rotmatrix, self._pos.T, order="F").T


class gauss(gnorm):
    """
    a gaussian shaped volume with random positions inside
    """

    def __init__(self, E, N, fwhm):
        super().__init__(E, N, fwhm, 2)


class multisphere(simobj):
    """
    multiple, randomly positioned spheres
    """

    @staticmethod
    @_numba.njit()
    def _staggeredadd(pos1, pos2, n):
        """
        repeat pos1[i] n[i] times and add it to pos2
        """
        pos2[: n[0], :] += pos1[0, :]
        for i in range(1, len(n)):
            pos2[n[i - 1] : min(len(pos2), n[i]), :] += pos1[i, :]
        pos2[n[-1] :, :] += pos1[-1, :]

    def __init__(self, E, Natoms=1e6, rsphere=10, fwhm=200, rho=2, spacing=1, Nspheres=_np.inf):
        """
        Multiple Spheres
        Parameters:
        Natoms: total number of excited atoms, if negative: mean number per sphere
        rsphere: radius of each sphere
        fwhm: the number of atoms per sphere scales generalised gaussian with the distance from the center. fhwm of the generalised gaussian.
        rho: parameter for generalised gaussian, rho=2 is gaussian in all directions.
        spacing: thickness of atom-free layer around each sphere
        Nspheres: max. total number of particles in volume
        """

        self._Nspheres, self.rsphere, self.spacing = Nspheres, rsphere, spacing
        self.fwhm, self.rho = _np.array(fwhm), _np.array(rho)
        self.rndPos = True
        self._debug = None
        N = int(-Natoms * _np.mean([len(self._spherepos()) for i in range(10)])) if Natoms < 0 else Natoms
        super().__init__(E, N)
        self._resetproperties = [
            "rndPos",
            "rsphere",
            "fwhm",
            "spacing",
            "rho",
            "_Nspheres",
        ]

    def _spherepos(self):
        return _pds(1.2 * self.fwhm, 2 * (self.rsphere + self.spacing), ndim=3, N=self._Nspheres)

    def updatePos(self):
        """
        calculate position of atoms
        """
        posspheres = self._spherepos()
        p = _np.exp(
            _ne.evaluate(
                "sum(-pos**rho*s, axis=1)",
                local_dict={
                    "s": _np.log(2) * (2 / self.fwhm) ** self.rho,
                    "pos": posspheres,
                    "rho": self.rho,
                },
            )
        )
        n = self.rng.poisson(p * (self._N / _np.sum(p)))
        missing = self._N - _np.sum(n)
        while missing:
            ids = self.rng.choice(len(n), int(abs(missing)), replace=True, p=p / (_np.sum(p)))
            _np.add.at(n, ids, _np.sign(missing))
            n[n < 0] = 0
            missing = self._N - _np.sum(n)
        nc = _np.cumsum(n)
        if self.rsphere > 0:
            posatoms = _rndSphere(self.rsphere, int(self._N))
        else:
            posatoms = _np.zeros((self._N, 3))
        multisphere._staggeredadd(posspheres, posatoms, nc)
        self._debug = (len(posspheres), _np.min(n), _np.max(n), _np.mean(n))
        self._pos = posatoms


class hcpsphere(multisphere):
    """
    densly hcp packed spheres
    """

    def __init__(
        self,
        E,
        Natoms=1e6,
        rsphere=10,
        fwhm=200,
        rho=2,
        a=20,
        rotangles=(0, 0, 0),
        sigma=0,
    ):
        self.Nhcp = None
        self.rndOrientation = False
        self.rsphere, self.a, self.sigma, self.rsphere = rsphere, a, sigma, rsphere
        self.fwhm, self.rho = _np.array(fwhm), _np.array(rho)
        self._rotmatrix = _rotation(*rotangles) if _np.any(rotangles) else False
        self._resetproperties = ["rsphere", "fwhm", "a", "sigma", "rho"]
        super().__init__(E, Natoms)

    def __setattr__(self, name, value):
        if any(name == n for n in ["sigma", "a", "fwhm"]):
            self._hcp = None
        super().__setattr__(name, value)

    def _rotate(self, rotmatrix):
        self._hcp = _np.matmul(rotmatrix, self._hcp.T, order="F").T
        self._pos = _np.matmul(rotmatrix, self._pos.T, order="F").T

    def _spherepos(self):
        if self._hcp is None or self.sigma:
            Nhcp = self.Nhcp or _np.ceil(5 * self.fwhm / (self.a * _np.array([2.0, 0.86, 1.63]))).astype(int)
            lconst = [self.a, self.a, 1.633 * self.a]
            unitcell = [[0, 0, 0], [1.0 / 3, 2.0 / 3, 1.0 / 2]]
            langle = _np.array([90, 90, 120]) * pi / 180.0
            hcp = crystal._lattice(lconst, langle, unitcell, Nhcp, self.sigma, self.rng)
            hcp -= _np.mean(hcp, axis=0)
            self._hcp = hcp

        if _np.any(self.rndOrientation):
            if _np.size(self.rndOrientation) == 3:
                rotmatrix = _axisrotation(
                    _np.array(self.rndOrientation, dtype=_np.float64),
                    self.rng.uniform(0, 2 * pi),
                )
            else:
                rotmatrix = _random_rotation(self.rng)
            if self._rotmatrix is not False:
                self._rotmatrix = rotmatrix @ self._rotmatrix
            else:
                self._rotmatrix = rotmatrix
        if self._rotmatrix is not False:
            self._hcp = _np.matmul(self._rotmatrix, self._hcp.T, order="F").T
        return self._hcp


class xyz(simobj):
    """
    atom positions specified by an xyz file
    """

    def __init__(self, E, filename, atomname, rotangles=(0, 0, 0), scale=1e-4):
        import re

        with open(filename, "r") as file:
            data = file.read()
        lines = re.findall(
            "^" + atomname + r"\d*\s*[\d,\.]+\s+[\d,\.]+\s+[\d,\.]+",
            data,
            re.IGNORECASE | re.MULTILINE,
        )
        pos = _np.genfromtxt(lines)[:, 1:] * scale
        self._pos = pos - (_np.max(pos, axis=0) / 2.0)
        self._rotmatrix = _rotation(*rotangles)
        self.rndOrientation = False
        if _np.any(rotangles):
            pos = _np.matmul(self._rotmatrix, pos.T, order="F").T
        super().__init__(E, len(pos))

    def _rotate(self, rotmatrix):
        self._pos = _np.matmul(rotmatrix, self._pos.T, order="F").T

    def updatePos(self):
        if self.rndOrientation:
            self._rotmatrix = _random_rotation(rng=self.rng)
            self._rotate(self._rotmatrix)


class crystal(simobj):
    """
    a crystalline structure
    """

    def __init__(
        self,
        E,
        lconst,
        langle,
        unitcell,
        N,
        repeats=None,
        rotangles=(0, 0, 0),
        fwhm=None,
        rho=2,
    ):
        if fwhm is not None:
            cosa, cosb, cosc = _np.cos(_np.array(langle))
            sina, sinb, sinc = _np.sin(_np.array(langle))
            neededrepeats = _np.ceil(
                2
                * _np.array(fwhm)
                / (
                    _np.array(
                        [
                            1,
                            sinc,
                            _np.sqrt(sinb**2 - ((cosa - cosb * cosc) / sinc) ** 2),
                        ]
                    )
                    * _np.array(lconst)
                )
            ).astype(int)
            if repeats is not None:
                if _np.any(neededrepeats > repeats):
                    _warn("Number of repeats small for choosen fwhm")
            else:
                repeats = neededrepeats
        if repeats is None:
            repeats = 3 * [int(_np.rint((N / len(unitcell)) ** (1 / 3.0)))]
        if _np.prod(repeats) * len(unitcell) < N:
            _warn("Number of atoms high for atoms in focus")
        allpos = crystal._lattice(lconst, langle, unitcell, repeats, rng=self.rng)

        if _np.any(rotangles):
            self._rotmatrix = _rotation(*rotangles)
            allpos = _np.matmul(self._rotmatrix, allpos.T, order="F").T
        else:
            self._rotmatrix = False

        self._allpos = allpos
        self.rndOrientation = False
        self._p = None
        self.fwhm = None if fwhm is None else _np.array(fwhm)
        self.rho = _np.array(2) if rho is None else _np.array(rho)

        super().__init__(E, N)
        self._resetproperties = ["rho", "fwhm"]

    @staticmethod
    def _lattice(lconst, langle, unitcell, repeats, sigma=0, rng=None):
        if rng is None:
            rng = _np.random.default_rng(_np.random.randint(2**63))
        cosa, cosb, cosc = _np.cos(_np.array(langle))
        sina, sinb, sinc = _np.sin(_np.array(langle))
        basis = _np.array(
            [
                [1, 0, 0],
                [cosc, sinc, 0],
                [
                    cosb,
                    (cosa - cosb * cosc) / sinc,
                    _np.sqrt(sinb**2 - ((cosa - cosb * cosc) / sinc) ** 2),
                ],
            ]
        ) * _np.expand_dims(lconst, 1)

        atoms = _np.dot(unitcell, basis).astype(_np.float32)
        if sigma != 0:
            atoms += rng.uniform(0, sigma, atoms.shape).astype(_np.float32)

        for j in range(3):
            atoms = _np.concatenate([atoms + (basis[j] * k).astype(_np.float32)[_np.newaxis, :] for k in range(repeats[j])])
            if sigma != 0:
                atoms += rng.uniform(0, sigma, atoms.shape).astype(_np.float32)

        return atoms - (_np.min(atoms, axis=0) + _np.max(atoms, axis=0)) / 2.0

    def updatePos(self):
        pos = self._allpos

        if _np.any(self.rndOrientation):
            if _np.size(self.rndOrientation) == 3:
                rotmatrix = _axisrotation(
                    _np.array(self.rndOrientation, dtype=_np.float64),
                    self.rng.uniform(0, 2 * pi),
                )
            else:
                rotmatrix = _random_rotation(self.rng)
            if self._rotmatrix is not False:
                self._rotmatrix = rotmatrix @ self._rotmatrix
            else:
                self._rotmatrix = rotmatrix
            pos = _np.matmul(self._rotmatrix, pos.T, order="F").T
            self._p = None

        if self.fwhm is not None and self._p is None:
            p = _np.exp(
                _ne.evaluate(
                    "sum(-pos**rho*s, axis=1)",
                    local_dict={
                        "s": _np.log(2) * (2 / self.fwhm) ** self.rho,
                        "pos": pos,
                        "rho": self.rho,
                    },
                )
            )
            p /= _np.sum(p)
            self._p = p
        else:
            p = self._p
        missing = self.N
        idx = []

        while missing:
            new = self.rng.choice(len(self._allpos), min(missing, len(self._allpos)), replace=False, p=p)
            idx.append(new)
            missing -= len(new)
        idx = _np.concatenate(idx)
        self._pos = pos[idx]

    def __setattr__(self, name, value):
        # reset cached values if property changes
        if any(name == n for n in self._resetproperties):
            self._p = None
        super().__setattr__(name, value)

    def _rotate(self, rotmatrix):
        self._allpos = _np.matmul(rotmatrix, self._allpos.T, order="F").T


class sc(crystal):
    """
    a sc crystal
    """

    def __init__(self, E, N, a, repeats=None, rotangles=(0, 0, 0), fwhm=None, rho=None):
        lconst = [a, a, a]
        unitcell = [[0, 0, 0]]
        langle = _np.array([90, 90, 90]) * pi / 180.0
        super().__init__(E, lconst, langle, unitcell, N, repeats, rotangles, fwhm, rho)


class fcc(crystal):
    """
    a fcc crystal
    """

    def __init__(self, E, N, a, repeats=None, rotangles=(0, 0, 0), fwhm=None, rho=None):
        lconst = [a, a, a]
        unitcell = [[0, 0, 0], [0.5, 0.5, 0], [0.5, 0, 0.5], [0, 0.5, 0.5]]
        langle = _np.array([90, 90, 90]) * pi / 180.0
        super().__init__(E, lconst, langle, unitcell, N, repeats, rotangles, fwhm, rho)


class hcp(crystal):
    """
    a hcp crystal
    """

    def __init__(self, E, N, a, repeats=None, rotangles=(0, 0, 0), fwhm=None, rho=None):
        lconst = [a, a, 1.633 * a]
        unitcell = [[0, 0, 0], [1.0 / 3, 2.0 / 3, 1.0 / 2]]
        langle = _np.array([90, 90, 120]) * pi / 180.0
        super().__init__(E, lconst, langle, unitcell, N, repeats, rotangles, fwhm, rho)


class cuso4(crystal):
    """
    cuso4 crsystal with fixed lattice constant
    """

    def __init__(self, N, E, repeats=None, rotangles=(0, 0, 0), fwhm=None, rho=None):
        # https://doi.org/10.1524%2Fzkri.1975.141.5-6.330
        unitcell = [[0, 0, 0], [0.5, 0.5, 0]]
        lconst = _np.array([6.141, 10.736, 5.986]) * 1e-4
        langle = _np.array([82.27, 107.43, 102.67]) * pi / 180.0
        super().__init__(E, lconst, langle, unitcell, N, repeats, rotangles, fwhm, rho)


class grating(simobj):
    def __init__(
        self,
        E,
        N,
        linewidth,
        spacingwidth,
        fwhm,
        rho=2,
        rholine=10,
        rotangles=(0, 0, 0),
    ):
        self.linewidth = linewidth
        self.spacingwidth = spacingwidth
        self.fwhm = fwhm
        self.rho = rho
        self.rholine = rholine
        self.__resetproperties = ["linewidth", "spacingwidth", "fwhm", "rho", "rholine"]
        if _np.any(rotangles):
            self._rotmatrix = _rotation(*rotangles)
        else:
            self._rotmatrix = False
        super().__init__(E, int(N))

    def updatePos(self):
        def _lines(x, a, b, rho):
            s = _np.log(2) * (2 / a) ** rho  # noqa
            return _ne.evaluate("exp(-abs((x%(a+b))-(a+b)/2)**rho*s)")

        if self._pos is None:
            rhofocusx = self.rho if _np.isscalar(self.rho) else self.rho[0]
            fwhmx = self.fwhm if _np.isscalar(self.fwhm) else self.fwhm[0]
            self._x = _np.arange(
                -(0.5 + 2 / rhofocusx) * fwhmx,
                (0.5 + 2 / rhofocusx) * fwhmx,
                min(self.linewidth, self.spacingwidth) / 10,
            )
            p = _lines(self._x, self.linewidth, self.spacingwidth, self.rholine) * _gnorm(self._x, fwhmx, rhofocusx)
            self._c = _np.cumsum(p)
            self._c = self._c / self._c[-1]
        rhofocusyz = self.rho if _np.isscalar(self.rho) else _np.array(self.rho[1:])
        fwhmyz = self.fwhm if _np.isscalar(self.fwhm) else _np.array(self.fwhm[1:])
        self._pos = _np.zeros((self.N, 3))
        self._pos[:, 0] = _np.interp(self.rng.uniform(size=self.N), self._c, self._x)
        self._pos[:, 1:] = _rndgennorm(0, fwhmyz, rhofocusyz, (self.N, 2), self.rng)

        if self._rotmatrix is not False:
            self._rotate(self._rotmatrix)

    def _rotate(self, rotmatrix):
        self._pos = _np.matmul(rotmatrix, self._pos.T, order="F").T


class membrane(simobj):
    """
    Tubes on a layer, such as a ALD coated AAO membrane
    """

    def __init__(
        self,
        E,
        poreradius=1,
        interporedistance=10,
        poreinnerradius=0,
        porelength=1,
        excitation=1,
        layerthickness=0,
        fwhm=200,
        rotangles=(0, 0, 0),
        rndOrientation=False,
    ):
        self._poreradius = poreradius
        self._poreinnerradius = poreinnerradius
        self._interporedistance = interporedistance
        self._porelength = porelength
        self._layerthickness = layerthickness
        self._fwhm = fwhm
        self._rndOrientation = rndOrientation

        if _np.any(rotangles):
            self._rotmatrix = _rotation(*rotangles)
        else:
            self._rotmatrix = False

        self._pores = self._getporepos(interporedistance, (4 * fwhm, 4 * fwhm))
        self._p = _np.exp(_np.sum(-(self._pores[:, :2] ** 2) * (4 * _np.log(2) / fwhm**2), axis=1))
        porevolume = porelength * _np.pi * (poreradius**2 - poreinnerradius**2)
        self._Natomspore = int(_np.sum(porevolume * excitation * self._p))
        self._Natomslayer = int(_np.pi * fwhm**2 / _np.log(16) * layerthickness * excitation)
        Natoms = int(self._Natomspore + self._Natomslayer)
        self._p /= _np.sum(self._p)

        super().__init__(E, Natoms)

    def updatePos(self):
        if self._pos is None:
            self._pos = _np.empty((self.N, 3))
        self._pos[: self._Natomspore, 1] = _np.sqrt(
            self.rng.random(self._Natomspore) * (self._poreradius**2 - self._poreinnerradius**2) + self._poreinnerradius**2
        )  # r
        self._pos[: self._Natomspore, 2] = self.rng.uniform(0, 2 * _np.pi, self._Natomspore)  # theta
        self._pos[: self._Natomspore, 0] = self._pos[: self._Natomspore, 1] * _np.cos(self._pos[: self._Natomspore, 2])
        self._pos[: self._Natomspore, 1] = self._pos[: self._Natomspore, 1] * _np.sin(self._pos[: self._Natomspore, 2])
        self._pos[: self._Natomspore, 2] = self.rng.uniform(
            self._layerthickness,
            self._layerthickness + self._porelength,
            self._Natomspore,
        )
        offset = self.rng.uniform(-self._interporedistance / 2, self._interporedistance / 2, 2).astype(_np.float32)
        self._p = _np.exp(
            _np.sum(
                -((self._pores[:, :2] - offset) ** 2) * (4 * _np.log(2) / self._fwhm**2),
                axis=1,
            )
        )
        self._p /= _np.sum(self._p)

        self._pos[: self._Natomspore, :] += self._pores[self.rng.choice(len(self._pores), self._Natomspore, p=self._p)]
        self._pos[self._Natomspore :, :2] = self.rng.normal(
            0,
            (self._fwhm / (_np.sqrt(2 * _np.log(2)) * 2)),
            size=(self._Natomslayer, 2),
        )
        self._pos[self._Natomspore :, 2] = self.rng.uniform(0, self._layerthickness, self._Natomslayer)
        self._pos[self._Natomspore :, :2] += offset

        if _np.any(self._rndOrientation):
            if _np.size(self._rndOrientation) == 3:
                rotmatrix = _axisrotation(
                    _np.array(self._rndOrientation, dtype=_np.float64),
                    self.rng.uniform(0, 2 * pi),
                )
            else:
                rotmatrix = _random_rotation(self.rng)
            if self._rotmatrix is not False:
                rotmatrix = self._rotmatrix @ rotmatrix
        else:
            rotmatrix = self._rotmatrix
        if rotmatrix is not False:
            _np.matmul(self._pos, rotmatrix.T, out=self._pos)

    def _getporepos(self, interporedistance, r):
        lconst = (interporedistance, interporedistance, 0)
        unitcell = (0, 0, 0)
        langle = _np.array((90, 90, 120)) * pi / 180.0
        repeats = (
            int(4 / _np.sqrt(3) * r[0] / interporedistance),
            int(4 / _np.sqrt(3) * r[1] / interporedistance),
            1,
        )
        pos = crystal._lattice(lconst, langle, unitcell, repeats)
        pos = pos[_np.linalg.norm(pos[:, :2] / _np.array(r), axis=1) < 1]
        return pos


class foil(simobj):
    """
    A foil with optional attenuation length
    """

    def __init__(
        self,
        E,
        excitation,
        thickness,
        fwhm,
        attenuationlength=_np.inf,
        rho=(2, 2),
        rotangles=(0, 0, 0),
    ):
        from math import log, exp
        from scipy.special import gamma

        self._rotmatrix = _rotation(*rotangles)
        self._rotangles = rotangles
        self._thickness = thickness
        self._attenuationlength = attenuationlength
        self._fwhm = (fwhm * _np.array((1, 1)))[:, None]
        self._rho = (rho * _np.array((1, 1)))[:, None]
        N = int(
            _np.prod(self._fwhm * gamma(1 + 1 / self._rho) * log(2) ** (-1 / self._rho))
            * (
                thickness / self._rotmatrix[-1, -1]
                if _np.isinf(attenuationlength)
                else (attenuationlength - attenuationlength / exp((thickness / self._rotmatrix[-1, -1]) / attenuationlength))
            )
            * excitation
        )
        super().__init__(E, N)

    def updatePos(self):
        if _np.all(self._rho == 2):
            self._pos = _np.matmul(
                self._rotmatrix[:, :-1],
                self.rng.normal(0, self._fwhm / 2.35, (2, self.N)),
                order="F",
            ).T
        else:
            self._pos = _np.matmul(
                self._rotmatrix[:, :-1],
                _rndgennorm(0, self._fwhm, self._rho, (2, self.N)),
                order="F",
            ).T
        if _np.isfinite(self.attenuationlength):
            self._pos[:, -1] += _rndtruncexp(
                self._attenuationlength,
                self._thickness / self._rotmatrix[-1, -1],
                self.N,
            )
        else:
            self._pos[:, -1] += self.rng.uniform(0, self._thickness / self._rotmatrix[-1, -1], self.N)

    # make everything readonly, as changing anything would change N
    @property
    def fwhm(self):
        return self._fwhm.ravel()

    @property
    def thickness(self):
        return self._thickness

    @property
    def rho(self):
        return self._rho.ravel()

    @property
    def attenuationlength(self):
        return self._attenuationlength

    @property
    def rotangles(self):
        return self._rotangles


class fake_view_from_position(simobj):
    """simobject that is rotated such that its cdi matches the idi from a different direction"""

    def __init__(self, simobj, detx, dety, detz):
        object.__setattr__(self, "_simobj", simobj)
        rotmatrix = _rotation(*_np.atan2(_np.array((-detx, dety, 0)), detz))
        object.__setattr__(self, "_rotmatrix", rotmatrix)

    def __getattr__(self, attr):
        print(attr)
        return getattr(self._simobj, attr)

    def __setattr__(self, name, value):
        setattr(self._simobj, "name", value)

    def get2(self):
        pos, phase = self._simobj.get2()
        pos = _np.matmul(self._rotmatrix, pos.T, order="F").T
        return pos, phase

    def updatePos(self):
        return self._simobj.updatePos()
